/*
 *
 *  Singe 2
 *  Copyright (C) 2019 Scott Duensing <scott@kangaroopunch.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *
 */


#ifndef UTIL_H
#define UTIL_H


#include "common.h"


char *utilCreateString(char *format, ...);
void  utilDie(char *fmt, ...);
bool  utilFileExists(char *filename);
char *utilGetFileExtension(char *filename);
char *utilGetLastPathComponent(char *pathname);
char  utilGetPathSeparator(void);
bool  utilPathExists(char *pathname);
void  utilSay(char *fmt, ...);


#endif // UTIL_H
